@@ -3,13 +3,21 @@ Reading settings from environment variables and providing a settings object
 for the application configuration.
 """
 
+import sys
+from io import StringIO
 from pathlib import Path
-from typing import Dict, List, Literal, Optional
+from typing import Dict, List, Literal, Optional, Set
+import threading
+import warnings
 
-from pydantic import BaseModel, ConfigDict, Field, field_validator
+from httpx import URL
+from pydantic import AliasChoices, BaseModel, ConfigDict, Field, field_validator
 from pydantic_settings import BaseSettings, SettingsConfigDict
 
 
+from mcp_agent.agents.agent_spec import AgentSpec
+
+
 class MCPServerAuthSettings(BaseModel):
     """Represents authentication configuration for a server."""
 
@@ -98,6 +106,10 @@ class MCPServerSettings(BaseModel):
     env: Dict[str, str] | None = None
     """Environment variables to pass to the server process."""
 
+    allowed_tools: Set[str] | None = None
+    """Set of tool names to allow from this server. If specified, only these tools will be exposed to agents. 
+    Tool names should match exactly. [WARNING] Empty list will result LLM have no access to tools."""
+
     model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
 
 
@@ -107,102 +119,273 @@ class MCPSettings(BaseModel):
     servers: Dict[str, MCPServerSettings] = Field(default_factory=dict)
     model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
 
+    @field_validator("servers", mode="before")
+    def none_to_dict(cls, v):
+        return {} if v is None else v
 
-class VertexAISettings(BaseModel):
-    """Settings for using VertexAI models in the MCP Agent application"""
 
-    project: str | None = None
-    location: str | None = None
+class VertexAIMixin(BaseModel):
+    """Common fields for Vertex AI-compatible settings."""
+
+    project: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("project", "PROJECT_ID", "GOOGLE_CLOUD_PROJECT"),
+    )
+
+    location: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "location", "LOCATION", "CLOUD_LOCATION", "GOOGLE_CLOUD_LOCATION"
+        ),
+    )
 
     model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
 
 
-class BedrockSettings(BaseModel):
+class BedrockMixin(BaseModel):
+    """Common fields for Bedrock-compatible settings."""
+
+    aws_access_key_id: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("aws_access_key_id", "AWS_ACCESS_KEY_ID"),
+    )
+
+    aws_secret_access_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("aws_secret_access_key", "AWS_SECRET_ACCESS_KEY"),
+    )
+
+    aws_session_token: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("aws_session_token", "AWS_SESSION_TOKEN"),
+    )
+
+    aws_region: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("aws_region", "AWS_REGION"),
+    )
+
+    profile: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("profile", "AWS_PROFILE"),
+    )
+
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
+
+class BedrockSettings(BaseSettings, BedrockMixin):
     """
     Settings for using Bedrock models in the MCP Agent application.
     """
 
-    aws_access_key_id: str | None = None
-    aws_secret_access_key: str | None = None
-    aws_session_token: str | None = None
-    aws_region: str | None = None
-    profile: str | None = None
-
-    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+    model_config = SettingsConfigDict(
+        env_prefix="",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
 
-class AnthropicSettings(VertexAISettings, BedrockSettings):
+class AnthropicSettings(BaseSettings, VertexAIMixin, BedrockMixin):
     """
     Settings for using Anthropic models in the MCP Agent application.
     """
 
-    api_key: str | None = None
-    default_model: str | None = None
-    provider: Literal["anthropic", "bedrock", "vertexai"] = "anthropic"
+    api_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "api_key", "ANTHROPIC_API_KEY", "anthropic__api_key"
+        ),
+    )
+    default_model: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "default_model", "ANTHROPIC_DEFAULT_MODEL", "anthropic__default_model"
+        ),
+    )
+    provider: Literal["anthropic", "bedrock", "vertexai"] = Field(
+        default="anthropic",
+        validation_alias=AliasChoices(
+            "provider", "ANTHROPIC_PROVIDER", "anthropic__provider"
+        ),
+    )
+    base_url: str | URL | None = Field(default=None)
 
-    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+    model_config = SettingsConfigDict(
+        env_prefix="ANTHROPIC_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
 
-class CohereSettings(BaseModel):
+class CohereSettings(BaseSettings):
     """
     Settings for using Cohere models in the MCP Agent application.
     """
 
-    api_key: str | None = None
+    api_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("api_key", "COHERE_API_KEY", "cohere__api_key"),
+    )
 
-    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+    model_config = SettingsConfigDict(
+        env_prefix="COHERE_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
 
-class OpenAISettings(BaseModel):
+class OpenAISettings(BaseSettings):
     """
     Settings for using OpenAI models in the MCP Agent application.
     """
 
-    api_key: str | None = None
-    reasoning_effort: Literal["low", "medium", "high"] = "medium"
-    base_url: str | None = None
-    user: str | None = None
+    api_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("api_key", "OPENAI_API_KEY", "openai__api_key"),
+    )
+
+    reasoning_effort: Literal["low", "medium", "high"] = Field(
+        default="medium",
+        validation_alias=AliasChoices(
+            "reasoning_effort", "OPENAI_REASONING_EFFORT", "openai__reasoning_effort"
+        ),
+    )
+    base_url: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "base_url", "OPENAI_BASE_URL", "openai__base_url"
+        ),
+    )
+
+    user: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices("user", "openai__user"),
+    )
 
     default_headers: Dict[str, str] | None = None
-    default_model: str | None = None
+    default_model: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "default_model", "OPENAI_DEFAULT_MODEL", "openai__default_model"
+        ),
+    )
 
     # NOTE: An http_client can be programmatically specified
     # and will be used by the OpenAI client. However, since it is
     # not a JSON-serializable object, it cannot be set via configuration.
     # http_client: Client | None = None
 
-    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+    model_config = SettingsConfigDict(
+        env_prefix="OPENAI_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
 
-class AzureSettings(BaseModel):
+class AzureSettings(BaseSettings):
     """
     Settings for using Azure models in the MCP Agent application.
     """
 
-    api_key: str | None = None
+    api_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "api_key", "AZURE_OPENAI_API_KEY", "AZURE_AI_API_KEY", "azure__api_key"
+        ),
+    )
 
-    endpoint: str
+    endpoint: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "endpoint", "AZURE_OPENAI_ENDPOINT", "AZURE_AI_ENDPOINT", "azure__endpoint"
+        ),
+    )
 
     credential_scopes: List[str] | None = Field(
         default=["https://cognitiveservices.azure.com/.default"]
     )
 
-    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+    model_config = SettingsConfigDict(
+        env_prefix="AZURE_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
 
-class GoogleSettings(BaseModel):
+class GoogleSettings(BaseSettings, VertexAIMixin):
     """
     Settings for using Google models in the MCP Agent application.
     """
 
-    api_key: str | None = None
-    """Or use the GOOGLE_API_KEY environment variable"""
+    api_key: str | None = Field(
+        default=None,
+        validation_alias=AliasChoices(
+            "api_key", "GOOGLE_API_KEY", "GEMINI_API_KEY", "google__api_key"
+        ),
+    )
 
-    vertexai: bool = False
+    vertexai: bool = Field(
+        default=False,
+        validation_alias=AliasChoices(
+            "vertexai", "GOOGLE_VERTEXAI", "google__vertexai"
+        ),
+    )
 
-    project: str | None = None
+    model_config = SettingsConfigDict(
+        env_prefix="GOOGLE_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
 
-    location: str | None = None
+
+class VertexAISettings(BaseSettings, VertexAIMixin):
+    """Standalone Vertex AI settings (for future use)."""
+
+    model_config = SettingsConfigDict(
+        env_prefix="VERTEXAI_",
+        extra="allow",
+        arbitrary_types_allowed=True,
+        env_file=".env",
+        env_file_encoding="utf-8",
+    )
+
+
+class SubagentSettings(BaseModel):
+    """
+    Settings for discovering and loading project/user subagents (AgentSpec files).
+    Supports common formats like Claude Code subagents.
+    """
+
+    enabled: bool = True
+    """Enable automatic subagent discovery and loading."""
+
+    search_paths: List[str] = Field(
+        default_factory=lambda: [
+            ".claude/agents",
+            "~/.claude/agents",
+            ".mcp-agent/agents",
+            "~/.mcp-agent/agents",
+        ]
+    )
+    """Ordered list of directories to scan. Earlier entries take precedence on name conflicts (project before user)."""
+
+    pattern: str = "**/*.*"
+    """Glob pattern within each directory to match files (YAML/JSON/Markdown supported)."""
+
+    definitions: List[AgentSpec] = Field(default_factory=list)
+    """Inline AgentSpec definitions directly in config."""
 
     model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
 
@@ -219,6 +402,15 @@ class TemporalSettings(BaseModel):
     task_queue: str
     max_concurrent_activities: int | None = None
     timeout_seconds: int | None = 60
+    rpc_metadata: Dict[str, str] | None = None
+    id_reuse_policy: Literal[
+        "allow_duplicate",
+        "allow_duplicate_failed_only",
+        "reject_duplicate",
+        "terminate_if_running",
+    ] = "allow_duplicate"
+
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
 
 
 class UsageTelemetrySettings(BaseModel):
@@ -233,6 +425,8 @@ class UsageTelemetrySettings(BaseModel):
     enable_detailed_telemetry: bool = False
     """If enabled, detailed telemetry data, including prompts and agents, will be sent to the telemetry server."""
 
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class TracePathSettings(BaseModel):
     """
@@ -257,15 +451,22 @@ class TracePathSettings(BaseModel):
     Uses Python's datetime.strftime format.
     """
 
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class TraceOTLPSettings(BaseModel):
     """
     Settings for OTLP exporter in OpenTelemetry.
     """
 
-    endpoint: str
+    endpoint: str | None = None
     """OTLP endpoint for exporting traces."""
 
+    headers: Dict[str, str] | None = None
+    """Optional headers for OTLP exporter."""
+
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class OpenTelemetrySettings(BaseModel):
     """
@@ -300,6 +501,8 @@ class OpenTelemetrySettings(BaseModel):
     Ignored if 'path' is specified.
     """
 
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class LogPathSettings(BaseModel):
     """
@@ -326,6 +529,8 @@ class LogPathSettings(BaseModel):
     Uses Python's datetime.strftime format.
     """
 
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class LoggerSettings(BaseModel):
     """
@@ -372,6 +577,8 @@ class LoggerSettings(BaseModel):
     http_timeout: float = 5.0
     """HTTP timeout seconds for event transport"""
 
+    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
+
 
 class Settings(BaseSettings):
     """
@@ -386,7 +593,7 @@ class Settings(BaseSettings):
         nested_model_default_partial_update=True,
     )  # Customize the behavior of settings here
 
-    mcp: MCPSettings | None = MCPSettings()
+    mcp: MCPSettings | None = Field(default_factory=MCPSettings)
     """MCP config, such as MCP servers"""
 
     execution_engine: Literal["asyncio", "temporal"] = "asyncio"
@@ -395,22 +602,22 @@ class Settings(BaseSettings):
     temporal: TemporalSettings | None = None
     """Settings for Temporal workflow orchestration"""
 
-    anthropic: AnthropicSettings | None = None
+    anthropic: AnthropicSettings | None = Field(default_factory=AnthropicSettings)
     """Settings for using Anthropic models in the MCP Agent application"""
 
-    bedrock: BedrockSettings | None = None
+    bedrock: BedrockSettings | None = Field(default_factory=BedrockSettings)
     """Settings for using Bedrock models in the MCP Agent application"""
 
-    cohere: CohereSettings | None = None
+    cohere: CohereSettings | None = Field(default_factory=CohereSettings)
     """Settings for using Cohere models in the MCP Agent application"""
 
-    openai: OpenAISettings | None = None
+    openai: OpenAISettings | None = Field(default_factory=OpenAISettings)
     """Settings for using OpenAI models in the MCP Agent application"""
 
-    azure: AzureSettings | None = None
+    azure: AzureSettings | None = Field(default_factory=AzureSettings)
     """Settings for using Azure models in the MCP Agent application"""
 
-    google: GoogleSettings | None = None
+    google: GoogleSettings | None = Field(default_factory=GoogleSettings)
     """Settings for using Google models in the MCP Agent application"""
 
     otel: OpenTelemetrySettings | None = OpenTelemetrySettings()
@@ -422,6 +629,15 @@ class Settings(BaseSettings):
     usage_telemetry: UsageTelemetrySettings | None = UsageTelemetrySettings()
     """Usage tracking settings for the MCP Agent application"""
 
+    agents: SubagentSettings | None = SubagentSettings()
+    """Settings for defining and loading subagents for the MCP Agent application"""
+
+    def __eq__(self, other):  # type: ignore[override]
+        if not isinstance(other, Settings):
+            return NotImplemented
+        # Compare by full JSON dump to avoid differences in internal field-set tracking
+        return self.model_dump(mode="json") == other.model_dump(mode="json")
+
     @classmethod
     def find_config(cls) -> Path | None:
         """Find the config file in the current directory or parent directories."""
@@ -434,24 +650,80 @@ class Settings(BaseSettings):
 
     @classmethod
     def _find_config(cls, filenames: List[str]) -> Path | None:
-        """Find the config file of one of the possible names in the current directory or parent directories."""
+        """Find a file by name in current, parents, and `.mcp-agent` subdirs, with home fallback.
+
+        Search order:
+          - For each directory from CWD -> root:
+              - <dir>/<filename>
+              - <dir>/.mcp-agent/<filename>
+          - Home-level fallback:
+              - ~/.mcp-agent/<filename>
+        Returns the first match found.
+        """
         current_dir = Path.cwd()
 
-        # Check current directory and parent directories
-        while current_dir != current_dir.parent:
+        # Check current directory and parent directories (direct and .mcp-agent subdir)
+        while True:
             for filename in filenames:
-                config_path = current_dir / filename
-                if config_path.exists():
-                    return config_path
+                direct = current_dir / filename
+                if direct.exists():
+                    return direct
+
+                mcp_dir = current_dir / ".mcp-agent" / filename
+                if mcp_dir.exists():
+                    return mcp_dir
+
+            if current_dir == current_dir.parent:
+                break
             current_dir = current_dir.parent
 
+        # Home directory fallback
+        try:
+            home = Path.home()
+            for filename in filenames:
+                home_file = home / ".mcp-agent" / filename
+                if home_file.exists():
+                    return home_file
+        except Exception:
+            pass
+
         return None
 
 
+class PreloadSettings(BaseSettings):
+    """
+    Class for preloaded settings of the MCP Agent application.
+    """
+
+    model_config = SettingsConfigDict(env_prefix="mcp_app_settings_")
+
+    preload: str | None = None
+    """ A literal YAML string to interpret as a serialized Settings model.
+    For example, the value given by `pydantic_yaml.to_yaml_str(settings)`.
+    Env Var: `MCP_APP_SETTINGS_PRELOAD`.
+    """
+
+    preload_strict: bool = False
+    """ Whether to perform strict parsing of the preload string.
+    If true, failures in parsing will raise an exception.
+    If false (default), failures in parsing will fall through to the default
+    settings loading.
+    Env Var: `MCP_APP_SETTINGS_PRELOAD_STRICT`.
+    """
+
+
 # Global settings object
 _settings: Settings | None = None
 
 
+def _clear_global_settings():
+    """
+    Convenience for testing - clear the global memoized settings.
+    """
+    global _settings
+    _settings = None
+
+
 def get_settings(config_path: str | None = None) -> Settings:
     """Get settings instance, automatically loading from config file if available."""
 
@@ -477,6 +749,29 @@ def get_settings(config_path: str | None = None) -> Settings:
 
     merged_settings = {}
 
+    preload_settings = PreloadSettings()
+    preload_config = preload_settings.preload
+    if preload_config:
+        try:
+            # Write to an intermediate buffer to force interpretation as literal data and not a file path
+            buf = StringIO()
+            buf.write(preload_config)
+            buf.seek(0)
+            yaml_settings = yaml.safe_load(buf) or {}
+
+            # Preload is authoritative: construct from YAML directly (no env overlay)
+            return Settings(**yaml_settings)
+        except Exception as e:
+            if preload_settings.preload_strict:
+                raise ValueError(
+                    "MCP App Preloaded Settings value failed validation"
+                ) from e
+            # TODO: Decide the right logging call here - I'm cautious that it's in a very central scope
+            print(
+                f"MCP App Preloaded Settings value failed validation: {e}",
+                file=sys.stderr,
+            )
+
     # Determine the config file to use
     if config_path:
         config_file = Path(config_path)
@@ -516,4 +811,16 @@ def get_settings(config_path: str | None = None) -> Settings:
 
     # No valid config found anywhere
     _settings = Settings()
+
+    # Thread-safety advisory: warn when using global singleton from non-main thread
+    if (
+        threading.current_thread() is not threading.main_thread()
+        and config_path is None
+    ):
+        warnings.warn(
+            "get_settings() returned the global Settings singleton on a non-main thread. "
+            "In multithreaded environments, prefer passing a Settings instance explicitly to MCPApp("
+            "settings=...) or provide a per-thread config_path to avoid cross-thread coupling.",
+            stacklevel=2,
+        )
     return _settings
