@@ -7,6 +7,7 @@ from typing import (
     Any,
     Dict,
     Generic,
+    Literal,
     Optional,
     TypeVar,
     TYPE_CHECKING,
@@ -15,17 +16,16 @@ from typing import (
 
 from pydantic import BaseModel, ConfigDict, Field
 from mcp_agent.core.context_dependent import ContextDependent
-from mcp_agent.executor.temporal import TemporalExecutor
-from mcp_agent.executor.temporal.workflow_signal import (
+from mcp_agent.executor.workflow_signal import (
+    Signal,
     SignalMailbox,
-    TemporalSignalHandler,
 )
-from mcp_agent.executor.workflow_signal import Signal
 from mcp_agent.logging.logger import get_logger
 
 if TYPE_CHECKING:
     from temporalio.client import WorkflowHandle
     from mcp_agent.core.context import Context
+    from mcp_agent.executor.temporal import TemporalExecutor
 
 T = TypeVar("T")
 
@@ -53,6 +53,8 @@ class WorkflowState(BaseModel):
 
 
 class WorkflowResult(BaseModel, Generic[T]):
+    # Discriminator to disambiguate from arbitrary dicts
+    kind: Literal["workflow_result"] = "workflow_result"
     value: Optional[T] = None
     metadata: Dict[str, Any] = Field(default_factory=dict)
     start_time: float | None = None
@@ -94,7 +96,9 @@ class Workflow(ABC, Generic[T], ContextDependent):
         ContextDependent.__init__(self, context=context)
 
         self.name = name or self.__class__.__name__
-        self._logger = get_logger(f"workflow.{self.name}")
+        # Bind workflow logger to the provided context so events can carry
+        # the current upstream_session even when emitted from background tasks.
+        self._logger = get_logger(f"workflow.{self.name}", context=context)
         self._initialized = False
         self._workflow_id = None  # Will be set during run_async
         self._run_id = None  # Will be set during run_async
@@ -214,6 +218,7 @@ class Workflow(ABC, Generic[T], ContextDependent):
         # Using __mcp_agent_ prefix to avoid conflicts with user parameters
         provided_workflow_id = kwargs.pop("__mcp_agent_workflow_id", None)
         provided_task_queue = kwargs.pop("__mcp_agent_task_queue", None)
+        workflow_memo = kwargs.pop("__mcp_agent_workflow_memo", None)
 
         self.update_status("scheduled")
 
@@ -225,12 +230,13 @@ class Workflow(ABC, Generic[T], ContextDependent):
                 self._run_id = str(self.executor.uuid())
         elif self.context.config.execution_engine == "temporal":
             # For Temporal workflows, we'll start the workflow immediately
-            executor: TemporalExecutor = self.executor
+            executor: "TemporalExecutor" = self.executor
             handle = await executor.start_workflow(
                 self.name,
                 *args,
                 workflow_id=provided_workflow_id,
                 task_queue=provided_task_queue,
+                workflow_memo=workflow_memo,
                 **kwargs,
             )
             self._workflow_id = handle.id
@@ -455,6 +461,122 @@ class Workflow(ABC, Generic[T], ContextDependent):
                     else:
                         cb(sig_obj)
 
+        @workflow.query(name="token_tree")
+        def _query_token_tree(self) -> str:
+            """Return a best-effort token usage tree string from the workflow process.
+
+            Notes:
+            - Queries must be deterministic and fast. We avoid awaiting any locks and read
+              the current in-memory snapshot. This may be slightly stale during execution
+              but is safe and sufficient for observability.
+            """
+            try:
+                counter = getattr(self.context, "token_counter", None)
+                if not counter:
+                    return "(no token usage)"
+                root = getattr(counter, "_root", None)
+                if not root:
+                    return "(no token usage)"
+                return root.format_tree()
+            except Exception:
+                return "(no token usage)"
+
+        @workflow.query(name="token_summary")
+        def _query_token_summary(self) -> Dict[str, Any]:
+            """Return a JSON-serializable token usage summary from the workflow process.
+
+            Structure:
+              {
+                "total_usage": {"total_tokens": int, "input_tokens": int, "output_tokens": int},
+                "total_cost": float,
+                "models": {
+                  "<model>(<provider>)" | "<model>": {
+                    "input_tokens": int,
+                    "output_tokens": int,
+                    "total_tokens": int,
+                    "cost": float,
+                    "provider": str | None
+                  }
+                },
+                "token_tree": str
+              }
+            """
+            summary: Dict[str, Any] = {
+                "total_usage": {
+                    "total_tokens": 0,
+                    "input_tokens": 0,
+                    "output_tokens": 0,
+                },
+                "total_cost": 0.0,
+                "models": {},
+                "token_tree": "(no token usage)",
+            }
+
+            try:
+                counter = getattr(self.context, "token_counter", None)
+                if not counter:
+                    return summary
+
+                # Build tree string from current root snapshot
+                root = getattr(counter, "_root", None)
+                if not root:
+                    return summary
+
+                summary["token_tree"] = root.format_tree()
+                agg = root.aggregate_usage()
+                summary["total_usage"] = {
+                    "input_tokens": int(agg.input_tokens),
+                    "output_tokens": int(agg.output_tokens),
+                    "total_tokens": int(agg.total_tokens),
+                }
+
+                # Derive model usage strictly from the current tree to avoid cross-run accumulation
+                from collections import defaultdict as _dd
+
+                model_nodes = _dd(list)  # type: ignore[var-annotated]
+                try:
+                    counter._collect_model_nodes(root, model_nodes)  # type: ignore[attr-defined]
+                except Exception:
+                    model_nodes = {}
+
+                total_cost = 0.0
+                for (model_name, provider), nodes in getattr(
+                    model_nodes, "items", lambda: []
+                )():
+                    total_input = 0
+                    total_output = 0
+                    for n in nodes:
+                        total_input += int(getattr(n.usage, "input_tokens", 0) or 0)
+                        total_output += int(getattr(n.usage, "output_tokens", 0) or 0)
+                    total_tokens = total_input + total_output
+
+                    cost = 0.0
+                    try:
+                        cost = float(
+                            counter.calculate_cost(
+                                model_name, total_input, total_output, provider
+                            )
+                        )
+                    except Exception:
+                        cost = 0.0
+                    total_cost += cost
+
+                    key = f"{model_name} ({provider})" if provider else model_name
+                    summary["models"][key] = {
+                        "input_tokens": total_input,
+                        "output_tokens": total_output,
+                        "total_tokens": total_tokens,
+                        "cost": cost,
+                        "provider": provider,
+                    }
+
+                summary["total_cost"] = total_cost
+            except Exception:
+                # Return whatever we have
+                pass
+
+            return summary
+
     async def get_status(self) -> Dict[str, Any]:
         """
         Get the current status of the workflow.
@@ -510,6 +632,71 @@ class Workflow(ABC, Generic[T], ContextDependent):
 
     # Static registry methods have been moved to the WorkflowRegistry class
 
+    async def get_token_node(self, return_all_matches: bool = False):
+        """Return this Workflow's token node(s) from the global counter."""
+        if not self.context or not getattr(self.context, "token_counter", None):
+            return [] if return_all_matches else None
+        counter = self.context.token_counter
+        if return_all_matches:
+            nodes = await counter.get_workflow_node(
+                name=self.name, return_all_matches=True
+            )
+            # Also support matching by IDs if present
+            if self.id:
+                nodes += await counter.get_workflow_node(
+                    workflow_id=self.id, return_all_matches=True
+                )
+            if self.run_id:
+                nodes += await counter.get_workflow_node(
+                    run_id=self.run_id, return_all_matches=True
+                )
+            return nodes
+        # Prefer run_id, then workflow_id, then name
+        if self.run_id:
+            node = await counter.get_workflow_node(run_id=self.run_id)
+            if node:
+                return node
+        if self.id:
+            node = await counter.get_workflow_node(workflow_id=self.id)
+            if node:
+                return node
+        return await counter.get_workflow_node(name=self.name)
+
+    async def get_token_usage(self):
+        """Return aggregated token usage for this Workflow (including children)."""
+        node = await self.get_token_node()
+        return node.get_usage() if node else None
+
+    async def get_token_cost(self) -> float:
+        """Return total cost for this Workflow (including children)."""
+        node = await self.get_token_node()
+        return node.get_cost() if node else 0.0
+
+    async def watch_tokens(
+        self,
+        callback,
+        *,
+        threshold: int | None = None,
+        throttle_ms: int | None = None,
+        include_subtree: bool = True,
+    ) -> str | None:
+        """Watch this Workflow's token usage. Returns a watch_id or None if not available."""
+        node = await self.get_token_node()
+        if not node:
+            return None
+        return await node.watch(
+            callback,
+            threshold=threshold,
+            throttle_ms=throttle_ms,
+            include_subtree=include_subtree,
+        )
+
+    async def format_token_tree(self) -> str:
+        node = await self.get_token_node()
+        if not node:
+            return "(no token usage)"
+        return node.format_tree()
+
     async def update_state(self, **kwargs):
         """Syntactic sugar to update workflow state."""
         for key, value in kwargs.items():
@@ -534,14 +721,84 @@ class Workflow(ABC, Generic[T], ContextDependent):
         self._logger.debug(f"Initializing workflow {self.name}")
 
         if self.context.config.execution_engine == "temporal":
-            if isinstance(self.executor.signal_bus, TemporalSignalHandler):
-                # Attach the signal handler to the workflow
-                self.executor.signal_bus.attach_to_workflow(self)
-            else:
-                self._logger.warning(
-                    "Signal handler not attached: executor.signal_bus is not a TemporalSignalHandler"
+            # Lazy import to avoid requiring Temporal unless engine is set to temporal
+            try:
+                from mcp_agent.executor.temporal.workflow_signal import (
+                    TemporalSignalHandler,
                 )
 
+                if isinstance(self.executor.signal_bus, TemporalSignalHandler):
+                    # Attach the signal handler to the workflow
+                    self.executor.signal_bus.attach_to_workflow(self)
+                else:
+                    self._logger.warning(
+                        "Signal handler not attached: executor.signal_bus is not a TemporalSignalHandler"
+                    )
+            except Exception:
+                self._logger.warning(
+                    "Signal handler not attached: Temporal support unavailable"
+                )
+
+            # Read memo (if any) and set gateway overrides on context for activities
+            try:
+                from temporalio import workflow as _twf
+
+                # Preferred API: direct memo mapping from Temporal runtime
+                memo_map = None
+                try:
+                    memo_map = _twf.memo()
+                except Exception:
+                    # Fallback to info().memo if available
+                    try:
+                        _info = _twf.info()
+                        memo_map = getattr(_info, "memo", None)
+                    except Exception:
+                        memo_map = None
+
+                if isinstance(memo_map, dict):
+                    gateway_url = memo_map.get("gateway_url")
+                    gateway_token = memo_map.get("gateway_token")
+
+                    self._logger.debug(
+                        f"Proxy parameters: gateway_url={gateway_url}, gateway_token={gateway_token}"
+                    )
+
+                    if gateway_url:
+                        try:
+                            self.context.gateway_url = gateway_url
+                        except Exception:
+                            pass
+                    if gateway_token:
+                        try:
+                            self.context.gateway_token = gateway_token
+                        except Exception:
+                            pass
+            except Exception:
+                # Safe to ignore if called outside workflow sandbox or memo unavailable
+                pass
+
+            # Expose a virtual upstream session (passthrough) bound to this run via activities
+            # This lets any code use context.upstream_session like a real session.
+            try:
+                from mcp_agent.executor.temporal.session_proxy import SessionProxy
+
+                upstream_session = getattr(self.context, "upstream_session", None)
+
+                if upstream_session is None:
+                    self.context.upstream_session = SessionProxy(
+                        executor=self.executor,
+                        context=self.context,
+                    )
+
+                    app = self.context.app
+                    if app:
+                        # Ensure the app's logger is bound to the current context with upstream_session
+                        if app._logger and hasattr(app._logger, "_bound_context"):
+                            app._logger._bound_context = self.context
+            except Exception:
+                # Non-fatal if context is immutable early; will be set after run_id assignment in run_async
+                pass
+
         self._initialized = True
         self.state.updated_at = datetime.now(timezone.utc).timestamp()
 
