@@ -20,6 +20,11 @@ from mcp.types import (
     CallToolResult,
     CreateMessageRequestParams,
     CreateMessageResult,
+    GetPromptResult,
+    ListPromptsResult,
+    ListResourcesResult,
+    ListToolsResult,
+    ReadResourceResult,
     SamplingMessage,
     TextContent,
     PromptMessage,
@@ -163,6 +168,11 @@ class RequestParams(CreateMessageRequestParams):
     This is used to stably identify the user in the LLM provider's logs.
     """
 
+    strict: bool = False
+    """
+    Whether models that support strict mode should strictly enforce the response schema.
+    """
+
 
 class AugmentedLLMProtocol(Protocol, Generic[MessageParamT, MessageT]):
     """Protocol defining the interface for augmented LLMs"""
@@ -229,6 +239,8 @@ class AugmentedLLM(ContextDependent, AugmentedLLMProtocol[MessageParamT, Message
 
     provider: str | None = None
     logger: Union["Logger", None] = None
+    # Suggested node type for token tracking for base LLMs
+    token_node_type: str = "llm"
 
     def __init__(
         self,
@@ -285,6 +297,16 @@ class AugmentedLLM(ContextDependent, AugmentedLLMProtocol[MessageParamT, Message
         self.model_selector = self.context.model_selector
         self.type_converter = type_converter
 
+    async def __aenter__(self):
+        if self.agent:
+            await self.agent.__aenter__()
+
+        return self
+
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        if self.agent:
+            await self.agent.__aexit__(exc_type, exc_val, exc_tb)
+
     @abstractmethod
     async def generate(
         self,
@@ -422,6 +444,8 @@ class AugmentedLLM(ContextDependent, AugmentedLLMProtocol[MessageParamT, Message
         last_message = await self.get_last_message()
         return self.message_param_str(last_message) if last_message else None
 
+    # region Agent / MCP convenience methods
+
     async def pre_tool_call(
         self, tool_call_id: str | None, request: CallToolRequest
     ) -> CallToolRequest | bool:
@@ -514,6 +538,38 @@ class AugmentedLLM(ContextDependent, AugmentedLLMProtocol[MessageParamT, Message
                     ],
                 )
 
+    async def list_tools(self, server_name: str | None = None) -> ListToolsResult:
+        """Call the underlying agent's list_tools method for a given server."""
+        return await self.agent.list_tools(server_name=server_name)
+
+    async def list_resources(
+        self, server_name: str | None = None
+    ) -> ListResourcesResult:
+        """Call the underlying agent's list_resources method for a given server."""
+        return await self.agent.list_resources(server_name=server_name)
+
+    async def read_resource(
+        self, uri: str, server_name: str | None = None
+    ) -> ReadResourceResult:
+        """Call the underlying agent's read_resource method for a given server."""
+        return await self.agent.read_resource(uri=uri, server_name=server_name)
+
+    async def list_prompts(self, server_name: str | None = None) -> ListPromptsResult:
+        """Call the underlying agent's list_prompts method for a given server."""
+        return await self.agent.list_prompts(server_name=server_name)
+
+    async def get_prompt(
+        self, name: str, server_name: str | None = None
+    ) -> GetPromptResult:
+        """Call the underlying agent's get_prompt method for a given server."""
+        return await self.agent.get_prompt(name=name, server_name=server_name)
+
+    async def close(self):
+        """Close underlying agent connections."""
+        await self.agent.close()
+
+    # endregion
+
     def message_param_str(self, message: MessageParamT) -> str:
         """Convert an input message to a string representation."""
         return str(message)
@@ -679,3 +735,83 @@ class AugmentedLLM(ContextDependent, AugmentedLLMProtocol[MessageParamT, Message
             identifier = str(self.context.executor.uuid())
 
         return f"{prefix}-{identifier}"
+
+    # region Token tracking
+
+    async def get_token_node(
+        self, return_all_matches: bool = False, node_type: str | None = None
+    ):
+        """Return this LLM's token node(s) from the global counter."""
+        if not self.context or not getattr(self.context, "token_counter", None):
+            return [] if return_all_matches else None
+        counter = self.context.token_counter
+        # Prefer explicit node_type, else default to this class's suggested node type
+        t = node_type or getattr(self, "token_node_type", None)
+        if return_all_matches:
+            if t == "llm":
+                return await counter.get_llm_node(self.name, return_all_matches=True)
+            if t == "agent":
+                return await counter.get_agent_node(self.name, return_all_matches=True)
+            # Fallback: gather both types
+            nodes = await counter.get_llm_node(self.name, return_all_matches=True)
+            nodes += await counter.get_agent_node(self.name, return_all_matches=True)
+            return nodes
+        else:
+            if t == "agent":
+                node = await counter.get_agent_node(self.name)
+                if node:
+                    return node
+            if t == "llm" or not t:
+                node = await counter.get_llm_node(self.name)
+                if node:
+                    return node
+            # Fallback try agent if not found
+            return await counter.get_agent_node(self.name)
+
+    async def get_token_usage(self, node_type: str | None = None):
+        """Return aggregated token usage for this LLM node (including children)."""
+        if not self.context or not getattr(self.context, "token_counter", None):
+            return None
+        counter = self.context.token_counter
+        t = node_type or getattr(self, "token_node_type", None)
+        if t == "agent":
+            return await counter.get_agent_usage(self.name)
+        if t == "llm":
+            return await counter.get_node_usage(self.name, "llm")
+        # Unknown type: try both
+        return await counter.get_node_usage(self.name)
+
+    async def get_token_cost(self, node_type: str | None = None) -> float:
+        """Return total cost for this LLM node (including children)."""
+        if not self.context or not getattr(self.context, "token_counter", None):
+            return 0.0
+        counter = self.context.token_counter
+        t = node_type or getattr(self, "token_node_type", None)
+        if t:
+            return await counter.get_node_cost(self.name, t)
+        return await counter.get_node_cost(self.name)
+
+    async def watch_tokens(
+        self,
+        callback,
+        *,
+        threshold: int | None = None,
+        throttle_ms: int | None = None,
+        include_subtree: bool = True,
+        node_type: str | None = None,
+    ) -> str | None:
+        """Watch this LLM's token usage. Returns a watch_id or None if not available."""
+        if not self.context or not getattr(self.context, "token_counter", None):
+            return None
+        counter = self.context.token_counter
+        t = node_type or getattr(self, "token_node_type", None) or "llm"
+        return await counter.watch(
+            callback=callback,
+            node_name=self.name,
+            node_type=t,
+            threshold=threshold,
+            throttle_ms=throttle_ms,
+            include_subtree=include_subtree,
+        )
+
+    # endregion
